from getpass import getpass
from typing import Optional
from os import environ
import subprocess


def check_tty_available() -> bool:
    try:
        tty = open("/dev/tty", "rb")
        tty.close()
        return True
    except:
        return False


def ask_pass_terminal(prompt: str) -> Optional[str]:
    """
    Ask password using the terminal
    """
    try:
        # TODO, add terminal notification?
        return getpass(prompt)
    except:
        return None


def ask_pass_program(prompt: str, program: Optional[str] = None) -> Optional[str]:
    if program is None:
        program = environ.get("MIMIKEEPASS_SSH_ASKPASS")
    if program is None:
        raise Exception("Missing program name")
    res = subprocess.run([program, prompt], capture_output=True)
    if res.returncode != 0:
        return None
    password = res.stdout.decode("UTF-8")
    if password.endswith("\n"):
        password = password[:-1]
    return password


def ask_pass(prompt: str, variable: str) -> Optional[str]:
    """
    Ask password (or something else)
    """

    program = environ.get(variable)
    require = environ.get(variable + "_REQUIRE")
    has_gui = environ.get("DISPLAY") is not None

    if program is None:
        return ask_pass_terminal(prompt)

    if require == "never":
        return ask_pass_terminal(prompt)

    if require == "prefer":
        if has_gui:
            return ask_pass_program(prompt, program=program)
        else:
            return ask_pass_terminal(prompt)

    if require == "force":
        return ask_pass_program(prompt, program=program)

    if not has_gui:
        return ask_pass_terminal(prompt)

    if check_tty_available():
        return ask_pass_terminal(prompt)
    else:
        return ask_pass_program(prompt, program=program)



from typing import Optional, cast, Dict, Any
from socket import socket, SOCK_STREAM, AF_UNIX
from os import environ

from mimikeepass.environ import resolve_socket_path
from mimikeepass.socket import JsonSocket, FramingSocket


class MimiKeepassClient:

    _vsock: JsonSocket

    def __init__(self, sock: Optional[socket] = None):
        if sock is not None:
            pass
        else:
            socket_path = resolve_socket_path()
            sock = socket(SOCK_STREAM, AF_UNIX)
            try:
                sock.connect(socket_path)
            except:
                sock.close()
                raise
        self._vsock = JsonSocket(FramingSocket(sock, separator=b"\0"))

    def get_password(self, title=None, url=None, username=None) -> Optional[str]:
        parameters = {}
        if title is not None:
            parameters["title"] = title
        if url is not None:
            parameters["url"] = url
        if username is not None:
            parameters["username"] = username

        self._vsock.send(
            {"method": "fr.urdhr.mimikeepass.GetEntry", "parameters": parameters}
        )
        response = self._vsock.recv()
        if not isinstance(response, dict):
            return None
        res = cast(Dict[str, Any], response).get("password")
        if not isinstance(res, str):
            return None
        return res

    def close(self) -> None:
        if self._vsock is not None:
            self._vsock.close()



from sys import stderr
from os import getuid, unlink, environ, chmod, getpid, stat, stat_result
from socket import socket, socketpair, SOCK_STREAM, AF_UNIX, SOL_SOCKET, SO_PEERCRED
from typing import Optional, List, Dict, Any, cast
import struct
import json
from threading import Thread, Lock
from selectors import DefaultSelector, EVENT_READ
from time import clock_gettime, CLOCK_MONOTONIC
from dataclasses import dataclass

import click
from pykeepass import PyKeePass
from pykeepass.exceptions import CredentialsError

from mimikeepass.environ import resolve_socket_path
from mimikeepass.askpass.ssh import ask_pass
from mimikeepass.socket import FramingSocket, JsonSocket


class KeePassDatabase:
    filename: str
    password: str
    database: Optional[PyKeePass]
    statinfo: Optional[stat_result]

    def __init__(self, filename: str, password: str):
        self.filename = filename
        self.password = password
        self.database = PyKeePass(filename, password=password)
        self.statinfo = stat(self.filename)

    def refresh(self):
        statinfo = stat(self.filename)
        if (
            self.database is None
            or self.statinfo is None
            or statinfo.st_dev != self.statinfo.st_dev
            or statinfo.st_ino != self.statinfo.st_ino
            or statinfo.st_mtime != self.statinfo.st_mtime
        ):
            try:
                self.database = PyKeePass(self.filename, password=self.password)
                self.statinfo = stat(self.filename)
            except:
                # TODO, prompt for password again?
                self.database = None
                self.statinfo = None


def open_keepass(filename: str) -> KeePassDatabase:
    error: Optional[BaseException] = None
    for i in range(3):
        password = ask_pass(
            f"Password for keepass file {filename}: ", variable="SSH_ASKPASS"
        )
        if password is None:
            continue
        try:
            kp = KeePassDatabase(filename, password=password)
            return kp
        except CredentialsError as e:
            error = e
            stderr.write(f"Invalid password for {filename}\n")
            continue
    assert error is not None
    raise error


@dataclass
class Entry:
    title: Optional[str] = None
    url: Optional[str] = None
    username: Optional[str] = None
    password: Optional[str] = None
    tags: Optional[List[str]] = None


class MimiKeepassDaemon:
    _kp: List[KeePassDatabase]
    _max_buffer_size = 4096
    _lock: Lock
    _connections_lock: Lock
    _connections: int
    _idle_duration: float
    _notify_read: socket
    _notify_write: socket
    _idle_timestamp: Optional[float]

    def __init__(self, kp: List[PyKeePass], idle: float = 0):
        self._kp = kp
        self._max_buffer_size = 4096
        self._lock = Lock()
        self._connections_lock = Lock()
        self._connections = 0
        self._idle_timestamp = clock_gettime(CLOCK_MONOTONIC)
        self._idle_duration = idle
        self._notify_read, self._notify_write = socketpair()
        self._notify_read.setblocking(False)

    def get_entry(
        self,
        username: Optional[str] = None,
        url: Optional[str] = None,
        uuid: Optional[str] = None,
        title: Optional[str] = None,
    ) -> Optional[Entry]:
        with self._lock:
            for kp in self._kp:
                kp.refresh()
                if kp.database is None:
                    continue
                args = {}
                if username is not None:
                    args["username"] = username
                if url is not None:
                    args["url"] = url
                if uuid is not None:
                    args["uuid"] = uuid
                if title is not None:
                    args["title"] = title
                entry = kp.database.find_entries(first=True, **args)
                if entry is None:
                    continue
                return Entry(
                    title=entry.deref("title"),
                    url=entry.deref("url"),
                    username=entry.deref("username"),
                    password=entry.deref("password"),
                    tags=entry.deref("tags"),
                )
        return None

    def handle_request(self, request: Dict[str, Any]) -> object:
        # TODO, handle exceptions
        method = request["method"]
        parameters = request["parameters"]
        if method == "fr.urdhr.mimikeepass.GetEntry":
            res = self.get_entry(
                username=parameters.get("username"),
                url=parameters.get("url"),
                uuid=parameters.get("uuid"),
                title=parameters.get("title"),
            )
            if res is None:
                return None
            else:
                return {
                    "title": res.title,
                    "url": res.url,
                    "username": res.username,
                    "password": res.password,
                    "tags": res.tags,
                }
        else:
            # TODO, handle this
            raise Exception("Invalid request")

    def serve_conn(self, conn: socket):
        vsock = JsonSocket(FramingSocket(conn, separator=b"\0"))
        try:
            while True:
                request = vsock.recv()
                if request is None or not isinstance(request, dict):
                    return
                oneway = request.get("oneway")
                response = self.handle_request(request)
                if not oneway:
                    conn.sendall(json.dumps(response).encode("UTF-8") + b"\0")
        finally:
            vsock.close()
            notify = False
            with self._connections_lock:
                self._connections -= 1
                if self._connections == 0:
                    self._idle_timestamp = clock_gettime(CLOCK_MONOTONIC)
                    notify = True
            if notify:
                self._notify_write.send(b"\0")

    def _get_timeout(self) -> Optional[float]:
        if self._idle_duration <= 0:
            return None
        with self._connections_lock:
            if self._idle_timestamp is None:
                return None
            else:
                idle_duration = clock_gettime(CLOCK_MONOTONIC) - self._idle_timestamp
                return self._idle_duration - idle_duration

    def accept_connections(self, socks: List[socket]):
        uid = getuid()

        selector = DefaultSelector()
        for sock in socks:
            sock.setblocking(False)
            selector.register(sock, EVENT_READ)

        selector.register(self._notify_read, EVENT_READ)

        while True:

            timeout = self._get_timeout()
            if timeout is not None and timeout <= 0:
                return

            for key, _ in selector.select(timeout):

                if key.fileobj == self._notify_read:
                    try:
                        while True:
                            self._notify_read.recv(1000)
                    except BlockingIOError as e:
                        pass
                    continue

                conn, addr = cast(socket, key.fileobj).accept()
                creds = conn.getsockopt(SOL_SOCKET, SO_PEERCRED, struct.calcsize("3i"))
                conn_pid, conn_uid, conn_gid = struct.unpack("3i", creds)
                if conn_uid != uid:
                    conn.close()
                    continue

                with self._connections_lock:
                    self._idle_timestamp = None
                    self._connections += 1

                thread = Thread(target=self.serve_conn, args=[conn])
                thread.run()


def get_listen_fds() -> Optional[int]:
    listen_pid_var = environ.get("LISTEN_PID")
    listen_pid = None if listen_pid_var is None else int(listen_pid_var)
    if listen_pid != getpid():
        return None
    listen_fds_var = environ.get("LISTEN_FDS")
    listen_fds = None if listen_fds_var is None else int(listen_fds_var)
    if listen_fds is None:
        return None
    del environ["LISTEN_FDS"]
    del environ["LISTEN_PID"]
    if "LISTEN_FDNAMES" in environ:
        del environ["LISTEN_FDNAMES"]
    return listen_fds


def serve(files: str, socket_path: Optional[str], idle: float = 0):
    if len(files) == 0:
        raise Exception("No files specified")

    try:
        daemon = MimiKeepassDaemon([open_keepass(file) for file in files], idle=idle)
    except CredentialsError as e:
        exit(1)

    listen_fds = get_listen_fds()
    socket_paths = []
    try:
        socks: List[socket] = []
        if listen_fds is not None:
            for i in range(listen_fds):
                sock = socket(fileno=i + 3)
                socks.append(sock)
        else:
            socket_path = resolve_socket_path(socket_path)

            sock = socket(SOCK_STREAM, AF_UNIX)
            sock.bind(socket_path)
            chmod(socket_path, 0o600)
            sock.listen(10)

            socket_paths.append(socket_path)
            socks.append(sock)

        daemon.accept_connections(socks)

    finally:
        for socket_path in socket_paths:
            try:
                unlink(socket_path)
            except:
                pass



from typing import Optional, TypeVar, Protocol
from socket import socket
import json


T = TypeVar("T")


class Socket(Protocol[T]):
    def send(self, message: T) -> None: ...

    def recv(self) -> Optional[T]: ...

    def close(self) -> None: ...


class FramingSocket(Socket[bytes]):
    _sock: socket
    _buffer: bytes
    _max_buffer_size = 4096
    _separator: bytes

    def __init__(self, sock: socket, separator: bytes):
        self._sock = sock
        self._buffer = b""
        self._separator = separator

    def recv(self) -> Optional[bytes]:
        if len(self._buffer) > self._max_buffer_size:
            raise IOError("Message too long")
        while True:
            i = self._buffer.find(self._separator)
            if i >= 0:
                res = self._buffer[:i]
                self._buffer = self._buffer[i + len(self._separator) :]
                return res
            chunk = self._sock.recv(4096)
            if not chunk:
                return None
            if len(self._buffer) + len(chunk) > self._max_buffer_size:
                raise IOError("Message too long")
            self._buffer = self._buffer + chunk

    def send(self, message: bytes):
        self._sock.sendall(message + self._separator)

    def close(self):
        self._sock.close()


class JsonSocket(Socket[object]):
    _sock: Socket[bytes]

    def __init__(self, sock: Socket[bytes]):
        self._sock = sock

    def recv(self) -> Optional[object]:
        message = self._sock.recv()
        if message is None:
            return None
        return json.loads(message.decode("UTF-8"))

    def send(self, object):
        self._sock.send(json.dumps(object).encode("UTF-8"))

    def close(self):
        self._sock.close()



from typing import Optional

import click

from mimikeepass.client import MimiKeepassClient
import mimikeepass.daemon


@click.group
def main(): ...


@main.command()
@click.argument("files", nargs=-1)
@click.option("--socket", "socket_path")
@click.option("--idle", type=float, default=0)
def serve(files: str, socket_path: Optional[str], idle: float = 0):
    mimikeepass.daemon.serve(files=files, socket_path=socket_path, idle=idle)


@main.command()
@click.option("--title", type=str, default=None)
@click.option("--url", type=str, default=None)
@click.option("--username", type=str, default=None)
def password(title=None, url=None, username=None) -> Optional[str]:
    """
    Get a password from mimikeepass server
    """
    client = MimiKeepassClient()
    try:
        res = client.get_password(title=title, url=url, username=username)
        if res is None:
            exit(1)
        print(res)
    finally:
        client.close()


if __name__ == "__main__":
    main()



#!/usr/bin/python3

import subprocess
from os import environ
from sys import argv
from re import compile
from dataclasses import dataclass
from typing import Optional

from mimikeepass.askpass.core import ask_pass
from mimikeepass.client import MimiKeepassClient


PASSWORD_AUTH_PROMPT_RE = compile("^([^\s]+)@([^@\s]+)'s password: $")


@dataclass
class PasswordAuthenticationRequest:
    username: str
    host: str


def parse_password_authentication_prompt(
    prompt: str,
) -> Optional[PasswordAuthenticationRequest]:
    match = PASSWORD_AUTH_PROMPT_RE.match(prompt)
    if match is None:
        return None
    return PasswordAuthenticationRequest(username=match.group(1), host=match.group(2))


def get_password_mimikeepass(title=None, url=None, username=None) -> Optional[str]:
    """
    Get a password from mimikeepassd
    """
    client = MimiKeepassClient()
    try:
        return client.get_password(title=title, url=url, username=username)
    finally:
        client.close()


def resolve_password(prompt: str) -> Optional[str]:

    if environ.get("SSH_ASKPASS_PROMPT"):
        return ask_pass(prompt, variable="MIMIKEEPASS_SSH_ASKPASS")

    password: Optional[str] = None

    req = parse_password_authentication_prompt(prompt)
    if req is not None:

        host = req.host
        username = req.username

        url = "ssh://" + host

        password = get_password_mimikeepass(url=url, username=username)
        if password is not None:
            return password

        # WALLIX WAB support,
        # Process username like target_user@???@target_hostname:SSH:???:real_user@auth_system
        if ":" in username:
            password = get_password_mimikeepass(
                url=url, username=username.split(":")[-1]
            )
            if password is not None:
                return password

        # Fallback to terminal or MIMIKEEPASS_ASKPASS:
        return ask_pass(prompt, variable="MIMIKEEPASS_SSH_ASKPASS")

    else:
        return ask_pass(prompt, variable="MIMIKEEPASS_SSH_ASKPASS")


def main():
    prompt = argv[1]
    password = resolve_password(prompt)
    if password:
        print(password)
    else:
        exit(1)


if __name__ == "__main__":
    main()



from os import environ, getuid
from typing import Optional


def get_runtime_directory() -> str:
    return environ.get("XDG_RUNTIME_DIR") or (f"/run/user/{getuid()}")


def resolve_socket_path(socket_path: Optional[str] = None) -> str:
    socket_path = (
        socket_path
        or environ.get("MIMIKEEPASS_SOCKET")
        or get_runtime_directory() + "/mimikeepass.varlink"
    )
    if "/" not in socket_path:
        socket_path = get_runtime_directory() + socket_path
    return socket_path
